part of mahg_essential_package;

abstract class IApiService {
  static Function(dynamic response)? unauthrizedCallback;

  Future<ProcessResult<Map>> httpGet(String url, {Map<String, String>? header});

  Future<ProcessResult<String>> httpGetString(String url,
      {Map<String, String>? header});

  Future<ProcessResult<dynamic>> httpGetDynamic(String url,
      {Map<String, String>? header});

  Future<ProcessResult<dynamic>> httpPost(String url, dynamic data,
      [Map<String, String>? header, bool asString = false]);
  Future<ProcessResult<dynamic>> httpPostEx(String url, dynamic data,
      [Map<String, String>? header, bool addToken = true]);

  Future<ProcessResult<dynamic>> httpPut(String url, Map data,
      {Map<String, String>? header});

  Future<ProcessResult<dynamic>> httpPutEx(String url, dynamic data,
      {Map<String, String>? header});

  Future<ProcessResult<dynamic>> httpPatch(String url, Map data,
      {Map<String, String>? header});

  Future<ProcessResult<dynamic>> httpPatchEx(String url, dynamic data,
      {Map<String, String>? header});

  Future<ProcessResult<dynamic>> httpDeleteDynamic(String url,
      {Map<String, String>? header, Map<String, dynamic>? body});

  Future<ProcessResult<dynamic>> httpDeleteDynamicEx(String url,
      {Map<String, String>? header});

  Future<ProcessResult<String>> httpPostFile(
      String url, Uint8List fileData, String name,
      [Map<String, String>? requestFields,
      Map<String, String>? header,
      bool isImage = true]);
}

class ApiService implements IApiService {
  /// backend token
  static String get token => ConfigService.token;

  static Map<String, String>? _baseHeader;
  Map<String, String> get baseHeader {
    if (_baseHeader == null) {
      _baseHeader = {
        'DeviceType': kIsWeb
            ? 'WebBrowser'
            : Platform.isAndroid
                ? 'Android'
                : Platform.isIOS
                    ? 'iOS'
                    : Platform.isWindows
                        ? 'Windows'
                        : Platform.isMacOS
                            ? 'MacOS'
                            : Platform.isLinux
                                ? 'Linux'
                                : Platform.isFuchsia
                                    ? 'Fuchsia'
                                    : 'Unknown',
      };
      if (!kIsWeb) {
        _baseHeader!['Accept-Encoding'] = 'gzip';
      }
    }
    return _baseHeader!;
  }

  /// call backend and return ProcessResult with Map
  Future<ProcessResult<Map>> httpGet(String url,
      {Map<String, String>? header}) async {
    try {
      var result = await httpGetString(url, header: header);
      if (!result.success)
        return ProcessResult(
          success: false,
          errorMessage: result.errorMessage,
        );
      var map = json.decode(result.data ?? '');
      if (kDebugMode) print('$url: ${result.data?.length} length');

      return ProcessResult(success: true, data: map);
    } catch (err, t) {
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }

  /// custom Encodeer for DateTime issue
  static dynamic customEncode(dynamic item) {
    if (item is DateTime) {
      return intl.DateFormat("yyyy-MM-ddTHH:mm:ss").format(item);
    }
    return item;
  }

  /// call backend and return ProcessResult with Map
  Future<ProcessResult<dynamic>> httpPost(String url, dynamic data,
      [Map<String, String>? header, bool asString = false]) async {
    try {
      var httpClient = http.Client();
      if (kDebugMode) print('POST: ${await _getFullUrl(url)}');

      if (asString)
        data =
            json.encode(data, toEncodable: customEncode).replaceAll('\"', '\'');
      var body = json.encode(data, toEncodable: customEncode);

      var headers = {'content-type': 'application/json'};
      for (var h in baseHeader.keys) headers[h] = baseHeader[h]!;
      if (token.isNotEmpty) headers['Authorization'] = 'Bearer $token';

      http.Response response = await httpClient.post(
        await _getFullUrl(url),
        headers: {...headers, ...?header},
        body: utf8.encode(body),
      );

      if (response.statusCode == 401 &&
          IApiService.unauthrizedCallback != null) {
        IApiService.unauthrizedCallback?.call(response.statusCode);
      }
      if (response.statusCode != 200) //&& response.statusCode !=307
      {
        print('statusCode: ${response.statusCode}');
        return ProcessResult(
            success: false,
            errorMessage: 'Result code = ${response.statusCode}');
      }
      var r = jsonDecode(response.body);
      httpClient.close();
      return parseResponse(r);
      // } else
      //   return ProcessResult(success: false, errorMessage: 'Result code = ${result.statusCode}');
    } catch (err, t) {
      print('xxxxxxxxxxxx ERROR: POST');
      print(err);
      print(t);
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }

  /// call backend and return ProcessResult with Map
  Future<ProcessResult<dynamic>> httpPut(String url, Map data,
      {Map<String, String>? header}) async {
    try {
      var httpClient = http.Client();

      if (kDebugMode) print('PUT: ${await _getFullUrl(url)}');

      var headers = {'content-type': 'application/json'};
      for (var h in baseHeader.keys) headers[h] = baseHeader[h]!;
      if (token.isNotEmpty) headers['Authorization'] = 'Bearer $token';

      http.Response response = await httpClient.put(
        await _getFullUrl(url),
        headers: {...headers, ...?header},
        body: json.encode(data, toEncodable: customEncode),
      );
      if (response.statusCode == 401 &&
          IApiService.unauthrizedCallback != null) {
        IApiService.unauthrizedCallback?.call(response.statusCode);
      }
      if (response.statusCode != 200) //&& response.statusCode !=307
      {
        print(response.statusCode);
        return ProcessResult(
            success: false,
            errorMessage: 'Result code = ${response.statusCode}');
      }

      String reply = response.body;

      httpClient.close();

      var r = jsonDecode(reply);
      return parseResponse(r);
      // } else
      //   return ProcessResult(success: false, errorMessage: 'Result code = ${result.statusCode}');
    } catch (err, t) {
      print('xxxxxxxxxxxx ERROR: POST');
      print(err);
      print(t);
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }

  /// call backend and return ProcessResult with Map
  Future<ProcessResult<dynamic>> httpPostEx(String url, dynamic data,
      [Map<String, String>? header, bool addToken = true]) async {
    try {
      if (kDebugMode) print('POST: ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var result = await httpClient.post(await _getFullUrl(url),
          body: json.encode(data, toEncodable: customEncode),
          headers: addToken
              ? {
                  'content-type': 'application/json',
                  ...baseHeader,
                  ...?header,
                  'Authorization': 'Bearer $token'
                }
              : {
                  'content-type': 'application/json',
                  ...baseHeader,
                });
      httpClient.close();
      if (result.statusCode != 200 &&
          result.statusCode != 201 &&
          result.statusCode != 400 &&
          result.statusCode != 422) {
        if (result.statusCode == 401 &&
            IApiService.unauthrizedCallback != null) {
          IApiService.unauthrizedCallback?.call(result.statusCode);
        }
        print('XXXXXXX  Response status code ${result.statusCode}');
        return ProcessResult(
            success: false, errorMessage: 'Result code = ${result.statusCode}');
      }

      String reply = result.body;
      var r = jsonDecode(reply);

      return parseResponse(r);
    } catch (err, t) {
      print('xxxxxxxxxxxx ERROR: POST');
      print(err);
      print(t);
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }

  /// call backend and return ProcessResult with Map
  Future<ProcessResult<dynamic>> httpPutEx(String url, dynamic data,
      {Map<String, String>? header}) async {
    try {
      if (kDebugMode) print('PUT: ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var result = await httpClient.put(await _getFullUrl(url),
          body: json.encode(data, toEncodable: customEncode),
          headers: {
            'content-type': 'application/json',
            ...baseHeader,
            ...?header,
            'Authorization': 'Bearer $token'
          });
      httpClient.close();
      if (result.statusCode != 200 &&
          result.statusCode != 201 &&
          result.statusCode != 422) //&& response.statusCode !=307
      {
        if (result.statusCode == 401 &&
            IApiService.unauthrizedCallback != null) {
          IApiService.unauthrizedCallback?.call(result.statusCode);
        }
        print(result.statusCode);
        return ProcessResult(
            success: false, errorMessage: 'Result code = ${result.statusCode}');
      }

      String reply = result.body;

      var r = jsonDecode(reply);
      return parseResponse(r);
      // } else
      //   return ProcessResult(success: false, errorMessage: 'Result code = ${result.statusCode}');
    } catch (err, t) {
      print('xxxxxxxxxxxx ERROR: POST');
      print(err);
      print(t);
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }

  /// call backend and return ProcessResult with Map
  @override
  Future<ProcessResult<dynamic>> httpGetDynamic(String url,
      {Map<String, String>? header}) async {
    try {
      if (kDebugMode)
        print('GET --------------------- ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var response = await httpClient.get(
        await _getFullUrl(url),
        headers: {
          'content-type': 'application/json',
          ...baseHeader,
          ...?header,
          'Authorization': 'Bearer $token'
        },
      );
      httpClient.close();
      if (response.statusCode != 200 &&
          response.statusCode != 201 &&
          response.statusCode != 422) {
        print('XXXXXXXXXXX GET status is ${response.statusCode} $url');
        if (response.statusCode == 401 &&
            IApiService.unauthrizedCallback != null) {
          IApiService.unauthrizedCallback?.call(response.statusCode);
        }
        return ProcessResult(
            success: false,
            errorMessage: 'Result code = ${response.statusCode}');
      } else if (kDebugMode) print('GET response $url');
      String reply = response.body;

      if (reply.length == 0)
        return ProcessResult(success: false, errorMessage: 'Result is empty');

      var r = jsonDecode(reply);

      if (r['success'] != true) {
        print(r);
        if (r['errorCodeString'] == 'InvalidAuthentication' &&
            IApiService.unauthrizedCallback != null) {
          IApiService.unauthrizedCallback?.call(r);
        }
      }
      return parseResponse(r);
    } catch (err, t) {
      print('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx ERROR: GET: $url');
      print(err);
      print(t);
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }

  /// call backend and return ProcessResult with Map
  Future<ProcessResult<dynamic>> httpDeleteDynamic(String url,
      {Map<String, String>? header, Map<String, dynamic>? body}) async {
    try {
      if (kDebugMode)
        print('DELETE --------------------- ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var response = await httpClient.delete(await _getFullUrl(url),
          headers: {
            'content-type': 'application/json',
            ...baseHeader,
            ...?header,
            'Authorization': 'Bearer $token'
          },
          body: (body != null
              ? utf8.encode(json.encode(body, toEncodable: customEncode))
              : null));
      httpClient.close();
      if (response.statusCode == 401 &&
          IApiService.unauthrizedCallback != null) {
        IApiService.unauthrizedCallback?.call(response.statusCode);
      }
      if (response.statusCode != 200)
        return ProcessResult(
            success: false,
            errorMessage: 'Result code = ${response.statusCode}');

      String reply = response.body;

      if (reply.length == 0)
        return ProcessResult(success: false, errorMessage: 'Result is empty');

      var r = jsonDecode(reply);

      if (r['success'] != true) print(r);
      return parseResponse(r);
    } catch (err, t) {
      print('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx   ERROR: DELETE: $url');
      print(err);
      print(t);
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }

  /// call backend and return ProcessResult with Map
  Future<ProcessResult<dynamic>> httpDeleteDynamicEx(String url,
      {Map<String, String>? header}) async {
    try {
      if (kDebugMode)
        print('DELETE --------------------- ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var headers = {'content-type': 'application/json'};
      for (var h in baseHeader.keys) headers[h] = baseHeader[h]!;
      if (token.isNotEmpty) headers['Authorization'] = 'Bearer $token';

      http.Response response = await httpClient
          .delete(await _getFullUrl(url), headers: {...headers, ...?header});
      if (response.statusCode == 401 &&
          IApiService.unauthrizedCallback != null) {
        IApiService.unauthrizedCallback?.call(response.statusCode);
      }
      if (response.statusCode != 200 &&
          response.statusCode != 201 &&
          response.statusCode != 422)
        return ProcessResult(
            success: false,
            errorMessage: 'Result code = ${response.statusCode}');

      String reply = response.body;
      httpClient.close();

      if (reply.length == 0)
        return ProcessResult(success: false, errorMessage: 'Result is empty');

      var r = jsonDecode(reply);

      if (r['success'] != true) print(r);
      return parseResponse(r);
    } catch (err, t) {
      print('xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx   ERROR: DELETE: $url');
      print(err);
      print(t);
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }

  @override
  Future<ProcessResult<String>> httpGetString(String url,
      {Map<String, String>? header}) async {
    try {
      var httpClient = http.Client();
      var result = await httpClient.get(await _getFullUrl(url), headers: {
        ...baseHeader,
        ...?header,
        'Authorization': 'Bearer '
            '$token'
      });
      httpClient.close();
      if (result.statusCode == 200) {
        return ProcessResult(success: true, data: result.body);
      } else {
        return ProcessResult(
            success: false, errorMessage: 'Result code = ${result.statusCode}');
      }
    } catch (err, t) {
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }

  static String? get apiURL => ConfigService.apiURL;
  static Future<Uri> _getFullUrl(String url) async {
    if (url.startsWith('http')) return Uri.parse(url);
    while (apiURL == null || (apiURL?.isEmpty ?? true)) {
      await Future.delayed(Duration(milliseconds: 100));
    }
    return Uri.parse("$apiURL$url");
  }

  /// call backend and post file like images
  @override
  Future<ProcessResult<String>> httpPostFile(
      String url, Uint8List fileData, String name,
      [Map<String, String>? requestFields,
      Map<String, String>? header,
      bool isImage = true]) async {
    String dataString;
    try {
      //print('POST:---- ${_getFullUrl(url)}   $name');
      //var httpClient = http.Client();
      var request = http.MultipartRequest("POST", await _getFullUrl(url));

      request.headers['Authorization'] = 'Bearer $token';
      // request.fields['user'] = 'someone@somewhere.com';
      request.files.add(http.MultipartFile.fromBytes(
        'file',
        fileData,
        filename: name,
        contentType: isImage
            ? MediaType("image", path.extension(name))
            : MediaType('application', 'x-tar'),
      ));
      // request.fields['name'] = name;
      request.fields['path'] = 'img';
      request.headers["Content-Type"] = "image/jpg";
      if (requestFields != null) {
        request.fields.addAll(requestFields);
      }
      if (header != null) {
        request.headers.addAll(header);
      }
      //request.headers["Content-Length"] = fileData.length.toString();
      var response = await request.send();
      if (response.statusCode == 200) {
        var data = (await response.stream.toBytes());
        if (kDebugMode) print(data.length);
        dataString = utf8.decode(data);
        var r = json.decode(dataString);
        return ProcessResult(
          data: r['data'] is Map ? r['data']['url'] : r['data'],
          success: r['success'] is bool
              ? r['success']
              : r['success']?.toString() == "true",
          errorMessage: r['errorMessage'] is Map || r['errorMessage'] == null
              ? (r['errorMessageAr'] ??
                  r['errorString'] ??
                  r['fullerrorMessagesString'] ??
                  r['errorCodeString'] ??
                  'No Server Error!')
              : r['errorMessage'],
        );
      }
      return ProcessResult(
          success: false,
          errorMessage: 'Error HttpPosFile: ${response.statusCode}');
    } catch (err, t) {
      print('XXXXXXXXXXXXXXXX post file error: $err');
      print(t);
      return ProcessResult(success: false, errorMessage: err.toString());
    }
  }

  ProcessResult parseResponse(Map r) {
    return ProcessResult(
        success: r['success'] is bool
            ? r['success']
            : r['success']?.toString() == "true",
        errorMessage: r['errorMessage'] ?? 'No Server Error!'

        // r['errorMessage'] is Map || r['errorMessage'] == null
        //     ? (r['errorMessageAr'] ??
        //         r['errorString'] ??
        //         r['fullerrorMessagesString'] ??
        //         r['errorCodeString'] ??
        //         'No Server Error!')
        //     : r['errorMessage']

        ,
        data: r['data'],
        token: r['token'],
        innerError: r['innerError']);
  }

  @override
  Future<ProcessResult> httpPatch(String url, Map data,
      {Map<String, String>? header}) async {
    try {
      var httpClient = http.Client();

      if (kDebugMode) print('PATCH: ${await _getFullUrl(url)}');

      var headers = {'content-type': 'application/json'};
      for (var h in baseHeader.keys) headers[h] = baseHeader[h]!;
      if (token.isNotEmpty) headers['Authorization'] = 'Bearer $token';

      http.Response response = await httpClient.patch(
        await _getFullUrl(url),
        headers: {...headers, ...?header},
        body: json.encode(data, toEncodable: customEncode),
      );

      if (response.statusCode == 401 &&
          IApiService.unauthrizedCallback != null) {
        IApiService.unauthrizedCallback?.call(response.statusCode);
      }
      if (response.statusCode != 200 && response.statusCode != 404) {
        print(response.statusCode);
        return ProcessResult(
            success: false,
            errorMessage: 'Result code = ${response.statusCode}');
      }

      String reply = response.body;

      httpClient.close();

      var r = jsonDecode(reply);
      return parseResponse(r);
      // } else
      //   return ProcessResult(success: false, errorMessage: 'Result code = ${result.statusCode}');
    } catch (err, t) {
      print('xxxxxxxxxxxx ERROR: POST');
      print(err);
      print(t);
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }

  @override
  Future<ProcessResult> httpPatchEx(String url, data,
      {Map<String, String>? header}) async {
    try {
      if (kDebugMode) print('PATCH: ${await _getFullUrl(url)}');
      var httpClient = http.Client();
      var result = await httpClient.put(await _getFullUrl(url),
          body: json.encode(data, toEncodable: customEncode),
          headers: {
            'content-type': 'application/json',
            ...baseHeader,
            ...?header,
            'Authorization': 'Bearer $token'
          });
      httpClient.close();

      if (result.statusCode != 200 &&
          result.statusCode != 201 &&
          result.statusCode != 422) //&& response.statusCode !=307
      {
        if (result.statusCode == 401 &&
            IApiService.unauthrizedCallback != null) {
          IApiService.unauthrizedCallback?.call(result.statusCode);
        }
        print(result.statusCode);
        return ProcessResult(
            success: false, errorMessage: 'Result code = ${result.statusCode}');
      }

      String reply = result.body;

      var r = jsonDecode(reply);
      return parseResponse(r);
      // } else
      //   return ProcessResult(success: false, errorMessage: 'Result code = ${result.statusCode}');
    } catch (err, t) {
      print('xxxxxxxxxxxx ERROR: POST');
      print(err);
      print(t);
      return ProcessResult(
        success: false,
        errorMessage: err.toString(),
      );
    }
  }
}
